#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

"""
IdaGeoDataFrame
"""

# Ensure Python 2 compatibility
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import super
from builtins import zip
from builtins import str
from builtins import int
from future import standard_library
standard_library.install_aliases()

import ibmdbpy4nps
from ibmdbpy4nps.frame import IdaDataFrame
from ibmdbpy4nps.geoSeries import IdaGeoSeries
from ibmdbpy4nps.exceptions import IdaGeoDataFrameError

from copy import deepcopy

import six


class IdaGeoDataFrame(IdaDataFrame):
    """  
    An IdaGeoDataFrame container inherits from IdaDataFrame.

    It has a property called "geometry" which refers to a column with
    geometry type. It is set as a string with a column name, either at
    instantiation time or with the set_geometry() method.

    If the "geometry" property is set, when calling a geospatial method from
    IdaDataFrame the method will be carried on the column this property refers
    to.

    The property "geometry" returns an IdaGeoSeries.

    See IdaDataFrame.
    See IdaGeoSeries.

    Notes
    -----
    IdaGeoDataFrame objects are not supported on Netezza.

    Examples
    --------
    >>> idageodf = IdaGeoDataFrame(idadb, 'SAMPLES.GEO_COUNTY',
    indexer='OBJECTID')
    >>> idageodf.dtypes
                     TYPENAME
    OBJECTID          INTEGER
    SHAPE     ST_MULTIPOLYGON
    STATEFP           VARCHAR
    COUNTYFP          VARCHAR
    COUNTYNS          VARCHAR
    NAME              VARCHAR
    GEOID             VARCHAR
    NAMELSAD          VARCHAR
    LSAD              VARCHAR
    CLASSFP           VARCHAR
    MTFCC             VARCHAR
    CSAFP             VARCHAR
    CBSAFP            VARCHAR
    METDIVFP          VARCHAR
    FUNCSTAT          VARCHAR
    ALAND             DECIMAL
    AWATER            DECIMAL
    INTPTLAT          VARCHAR
    INTPTLON          VARCHAR

    >>> idageodf[['NAME', 'SHAPE']].head()
           NAME                                              SHAPE
    0    Becker  MULTIPOLYGON (((-95.1637185512 46.7176480983, ...
    1  Jim Hogg  MULTIPOLYGON (((-98.9542377853 26.7856984795, ...
    2     Henry  MULTIPOLYGON (((-88.0532984194 36.4970648458, ...
    3     Keith  MULTIPOLYGON (((-102.0517705602 41.0038968011,...
    4   Clinton  MULTIPOLYGON (((-94.2059683962 39.7458481141, ...

    >>> idageodf.geometry
    AttributeError: Geometry property has not been set yet. Use set_geometry
    method to set it.

    >>> idageodf.set_geometry('SHAPE')
    >>> idageodf.geometry.column
    'SHAPE'

    >>> type(idageodf.geometry)
    <class 'ibmdbpy.geoSeries.IdaGeoSeries'>

    >>> idageoseries = idageodf.geometry    
    >>> idageoseries.head()
    0    MULTIPOLYGON (((-95.1637185512 46.7176480983, ...
    1    MULTIPOLYGON (((-98.9542377853 26.7856984795, ...
    2    MULTIPOLYGON (((-88.0532984194 36.4970648458, ...
    3    MULTIPOLYGON (((-102.0517705602 41.0038968011,...
    4    MULTIPOLYGON (((-94.2059683962 39.7458481141, ...
    Name: SHAPE, dtype: object

    >>> idageodf['County area'] = idageodf.area(unit='mile')

    >>> counties_with_areas = idageodf[['NAME', 'SHAPE', 'County area']]
    >>> counties_with_areas.dtypes
                        TYPENAME
    NAME                 VARCHAR
    SHAPE        ST_MULTIPOLYGON
    County area           DOUBLE

    >>> counties_with_areas.head()
            NAME                                             SHAPE  County area
    0     Menard  MULTIPOLYGON (((-99.4847630885 30.940610279, ...   902.281540
    1      Boone  MULTIPOLYGON (((-88.7764991497 42.491919892, ...   282.045087
    2  Ochiltree  MULTIPOLYGON (((-100.5467326897 36.056542135,...   918.188142
    3    Sharkey  MULTIPOLYGON (((-90.9143429922 33.007703026, ...   435.548518
    4    Audubon  MULTIPOLYGON (((-94.7006367168 41.504155369, ...   444.827726
    """

    def __init__(self, idadb, tablename, indexer = None, geometry = None):
        """
        Constructor for IdaGeoDataFrame objects.
        See IdaDataFrame.__init__ documentation.

        Parameters
        ----------
        geometry : str, optional
            Column name to set the "geometry" property of the IdaGeoDataFrame.
            The column must have geometry type.

        Attributes
        ----------
        _geometry_colname : str
            Name of the column that "geometry" property refers to.
            This attribute must be set through the set_geometry() method.
        geometry : IdaGeoSeries
            The column referenced by _geometry_colname attribute.
        """
        # TODO: Add support for receiving either a string or an IdaGeoSeries as 
        # geometry parameter.        

        if (idadb.__class__.__name__ == "IdaDataBase") & idadb._is_netezza_system():
            raise IdaGeoDataFrameError("IdaGeoDataFrame objects are not supported on Netezza.")

        if geometry is not None and not isinstance(geometry, six.string_types):
            raise TypeError("geometry must be a string")
        super(IdaGeoDataFrame, self).__init__(idadb, tablename, indexer)
        self._geometry_colname = None
        if geometry is not None:
            self.set_geometry(geometry)

    def __getitem__(self, item):
        """
        Returns an IdaDataFrame, IdaSeries, IdaGeoDataFrame or IdaGeoSeries
        as appropriate.
        
        Returns
        --------
        IdaGeoSeries
            When the projection has only one column and it has geometry type.
        IdaGeoDataFrame
            When the projection has more than one column, and the "geometry"
            column of the IdaGeoDataFrame is included in them.
        IdaDataFrame
            When the projection has more than one column, and the "geometry"
            column of the IdaGeoDataFrame is not included in them.
        IdaSeries
            When the projection has only one column and it doesn't have 
            geometry type.
        """
        ida = super(IdaGeoDataFrame, self).__getitem__(item)
        if isinstance(ida, ibmdbpy4nps.IdaSeries):
            if ida.dtypes['TYPENAME'][ida.column].find('ST_') == 0:
                idageoseries = IdaGeoSeries.from_IdaSeries(ida)
                # Return IdaGeoSeries
                return idageoseries
            else:
                # Return IdaSeries
                return ida
        elif isinstance(ida, ibmdbpy4nps.IdaDataFrame):
            if self._geometry_colname in ida.dtypes.index:
                # Return IdaGeoDataFrame
                idageodf = IdaGeoDataFrame.from_IdaDataFrame(ida)
                idageodf._geometry_colname = self._geometry_colname
                return idageodf
            else:
                # Return IdaDataFrame
                return ida

    def __delitem__(self, item):
        """
        Erases the "geometry" property if the column it refers to is deleted.
        """
        super(IdaGeoDataFrame, self).__delitem__(item)
        if item == self._geometry_colname:
            self._geometry_colname = None

    def __getattr__(self, name):
        """
        Carry geospatial method calls on the "geometry" column of the
        IdaGeoDataFrame, if it was set.
        
        Notes
        -----
        This method gets called only when an attribute lookup on
        IdaGeoDataFrame is not resolved, i.e. it is not an instance attribute
        and it's not found in the class tree.
        """        
        
        if name == 'geometry':
            # When .geometry is accessed and _geometry_colname is None
            return self.__getattribute__('geometry')

        if hasattr(IdaGeoSeries, name):
            # Geospatial method call
            if self._geometry_colname is None:
                raise AttributeError("Geometry column has not been set yet.")
            else:
                # Get a IdaGeoSeries and carry the operation on it
                idageoseries = self.__getitem__(item = self._geometry_colname)
                return idageoseries.__getattribute__(name)
        else:
            raise AttributeError
    
    @property
    def geometry(self):
        """
        Returns an IdaGeoSeries with the column whose name is stored in 
        _geometry_colname attribute.

        The setter calls the set_geometry() method.

        Returns
        -------
        IdaGeoSeries

        Raises
        ------
        AttributeError
            If the property has not been set yet.
        
        """
        if self._geometry_colname is None:
            raise AttributeError(
                "Geometry property has not been set yet. "
                "Use set_geometry method to set it.")
        else:
            return self.__getitem__(self._geometry_colname)
    
    @geometry.setter
    def geometry(self, value):
        """
        See set_geometry() method.
        """
        self.set_geometry(value)

    @classmethod
    def from_IdaDataFrame(cls, idadf, geometry = None):
        """ 
        Creates an IdaGeoDataFrame from an IdaDataFrame.
        
        Parameters
        ----------
        geometry : str, optional
            Column name to set the "geometry" property of the IdaGeoDataFrame.
            The column must have geometry type.

        Raises
        ------
        TypeError
            If idadf is not an IdaDataFrame.
        """
        
        if not isinstance(idadf, IdaDataFrame):
            raise TypeError("Expected IdaDataFrame")
        else:
            # TODO: check if it's better to only change the .__base__ attribute

            #behavior based on _clone() method of IdaDataFrame
            newida = IdaGeoDataFrame(
                    idadf._idadb, idadf._name, idadf.indexer, geometry)
            newida.columns = idadf.columns 
            newida.dtypes = idadf.dtypes
            
            newida.internal_state.name = deepcopy(idadf.internal_state.name)
            newida.internal_state.ascending = deepcopy(idadf.internal_state.ascending)
            #newida.internal_state.views = deepcopy(idadf.internal_state.views)
            newida.internal_state._views = deepcopy(idadf.internal_state._views)
            newida.internal_state._cumulative = deepcopy(idadf.internal_state._cumulative)
            newida.internal_state.order = deepcopy(idadf.internal_state.order)
            newida.internal_state.columndict = deepcopy(idadf.internal_state.columndict)
            return newida

    def set_geometry(self, column_name):
        """
        Receives a column name to set as the "geometry" column of the
        IdaDataFrame.

        Parameters
        -----------
        column_name : str
            Name of the column to be set as geometry column of the 
            IdaDataFrame. It must have geometry type.

        Raises
        ------
        KeyError
            If the column is not present in the IdaGeoDataFrame.
        TypeError
            If the column doesn't have geometry type.
        """
        if not isinstance(column_name, six.string_types):
            raise TypeError("column_name must be a string")
        if column_name not in self.columns:
            raise KeyError(
                "'" + column_name + "' cannot be set as geometry column: "
                "not a column in the IdaGeoDataFrame."
            )
        elif self.dtypes.TYPENAME[column_name].find('ST_') != 0:
            raise TypeError(
                "'" + column_name + "' cannot be set as geometry column: "
                "column doesn't have geometry type."
            )
        else:
            self._geometry_colname = column_name

    # ==============================================================================
    ### Binary geospatial methods
    # ==============================================================================
    def equals(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the geometry of the first IdaGeoDataFrame
        crosses the second.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.


        References
        ----------
        DB2 Spatial Extender ST_CROSSES() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.equals(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          0
        2            1840         0
        2            109          0
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_EQUALS',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def distance(self, ida2, unit=None):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with a numeric value
        which is the geographic distance measured between the
        geometries of the input IdaGeoDataFrames.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.
        unit : str, optional
            Name of the unit, it is case-insensitive.
            If omitted, the following rules are used:

                * If geometry is in a projected or geocentric coordinate
                  system, the linear unit associated with this coordinate system
                  is used.
                * If geometry is in a geographic coordinate system, the angular
                  unit associated with this coordinate system is used.

        References
        ----------
        DB2 Spatial Extender ST_DISTANCE() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.distance(ida2,unit = 'KILOMETER')
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          26.918942
        2            1840         4.868971
        2            109          16.387094
        """
        add_args = None
        if unit is not None:
            unit = self._check_linear_unit(unit)  # Can raise exceptions
            add_args = []
            add_args.append(unit)
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_DISTANCE',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'],
            additional_args = add_args)

    def crosses(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the geometry of the first IdaGeoDataFrame
        crosses the second.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_CROSSES() function.

        See also
        --------
        linear_units : list of valid units.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.crosses(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          0
        2            1840         0
        2            109          0
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_CROSSES',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def intersects(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the geometries of the input IdaGeoDataFrames
        intersect each other.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_INTERSECTS() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.intersects(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          0
        2            1840         0
        2            109          0
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_INTERSECTS',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def overlaps(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the geometries of the input IdaGeoDataFrames
        overlap each other.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation


        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_OVERLAPS() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.overlaps(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          0
        2            1840         0
        2            109          0
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_OVERLAPS',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def touches(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the boundary of the first geometry touches
        the second.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_TOUCHES() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.touches(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          0
        2            1840         0
        2            109          0
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_TOUCHES',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def disjoint(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the geometries in the input dataframes are
        disjoint.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_DISJOINT() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.disjoint(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          1
        2            1840         1
        2            109          1
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_DISJOINT',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def contains(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the second geometry contains the first.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_CONTAINS() function.

        Examples
        --------
        >>> idageodf_customer = IdaGeoDataFrame(idadb,'SAMPLES.GEO_CUSTOMER',indexer='OBJECTID')
        >>> idageodf_customer.set_geometry('SHAPE')
        >>> idageodf_county = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> idageodf_county.set_geometry('SHAPE')
        >>> ida1 = idageodf_customer[idageodf_customer['INSURANCE_VALUE']>250000]
        >>> ida2 = idageodf_county[idageodf_county['NAME']=='Madison']
        >>> result = ida2.contains(ida1)
        >>> result[result['RESULT']==1].head()
        INDEXERIDA1    INDEXERIDA2    RESULT
        21473          134            1
        21413          134            1
        21414          134            1
        21417          134            1
        21419          134            1
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_CONTAINS',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def within(self, ida2):
        """
        Valid types for the column in the calling IdaGeoDataFrame:
        ST_Geometry or one of its subtypes.

        Returns an IdaGeoDataFrame of indices of the two input
        IdaGeoDataFrames and a result column with 1 or 0 depending
        upon whether the first geometry is inside the second.

        For None geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_WITHIN() function.

        Examples
        --------
        >>> idageodf_customer = IdaGeoDataFrame(idadb,'SAMPLES.GEO_CUSTOMER',indexer='OBJECTID')
        >>> idageodf_customer.set_geometry('SHAPE')
        >>> idageodf_county = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> idageodf_county.set_geometry('SHAPE')
        >>> ida1 = idageodf_customer[idageodf_customer['INSURANCE_VALUE']>250000]
        >>> ida2 = idageodf_county[idageodf_county['NAME']=='Madison']
        >>> result = ida1.within(ida2)
        >>> result[result['RESULT']==1].head()
        INDEXERIDA1    INDEXERIDA2    RESULT
        134            21473          1
        134            21413          1
        134            21414          1
        134            21417          1
        134            21419          1
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_WITHIN',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def mbr_intersects(self, ida2):
        """
        This method takes a second IdaGeoDataFrame an an input
        and checks if the Minimum Bounding rectangles of the
        geometries from both IdaGeoDataFrames intersect and
        stores the result as 0 or 1 in the RESULT column of
        the resulting IdaGeoDataFrame.

        For None geometries the output is None.
        For empty geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_MBRIntersects() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.difference(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          0
        2            1840         0
        2            109          0
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_MBRINTERSECTS',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def difference(self, ida2):
        """
        This method takes a second IdaGeoDataFrame an an input
        and returns the difference of the geometries from both
        IdaGeoDataFrames as a new geometry stored in the RESULT
        column of the resulting IdaGeoDataFrame.

        For None geometries the output is None.
        For empty geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_Difference() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.difference(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          POLYGON ((-96.6219873342 30.0442882117, -96.61...
        2            1840         POLYGON ((-96.6219873342 30.0442882117, -96.61...
        2            109          POLYGON ((-96.6219873342 30.0442882117, -96.61...
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_DIFFERENCE',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def intersection(self, ida2):
        """
        This method takes a second IdaGeoDataFrame an an input
        and returns the intersection of the geometries from both
        IdaGeoDataFrames as a new geometry stored in the RESULT
        column of the resulting IdaGeoDataFrame.

        For None geometries the output is None.
        For empty geometries the output is POINT EMPTY.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_Intersection() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.intersection(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          POINT EMPTY
        2            1840         POINT EMPTY
        2            109          POINT EMPTY
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_INTERSECTION',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def union(self, ida2):
        """
        This method takes a second IdaGeoDataFrame an an input
        and returns the union of the geometries from both
        IdaGeoDataFrames as a new geometry stored in the RESULT
        column of the resulting IdaGeoDataFrame.

        For None geometries the output is None.
        For empty geometries the output is None.

        Returns
        -------
        Returns an IdaGeoDataFrame with three columns:

        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation

        Parameters
        ----------
        ida2 : IdaGeoDataFrame
            Name of the second IdaGeoDataFrame on which the function ST_EQUALS()
            will be invoked.

        References
        ----------
        DB2 Spatial Extender ST_Union() function.

        Examples
        --------
        >>> counties = IdaGeoDataFrame(idadb,'SAMPLES.GEO_COUNTY',indexer='OBJECTID')
        >>> counties.set_geometry('SHAPE')
        >>> ida1 = counties[counties['NAME'] == 'Austin']
        >>> ida2 = counties[counties['NAME'] == 'Kent']
        >>> result = ida1.union(ida2)
        >>> result.head()
        INDEXERIDA1  INDEXERIDA2  RESULT
        2            163          MULTIPOLYGON (((-96.6219873342 30.0442882117, ...
        2            1840         MULTIPOLYGON (((-96.6219873342 30.0442882117, ...
        2            109          MULTIPOLYGON (((-96.6219873342 30.0442882117, ..
        """
        return self._binary_operation_handler(
            ida2,
            db2gse_function='DB2GSE.ST_UNION',
            valid_types_ida1=['ST_GEOMETRY'],
            valid_types_ida2=['ST_GEOMETRY'])

    def _binary_operation_handler(self, ida2, db2gse_function,
                                          valid_types_ida1, valid_types_ida2,
                                          additional_args=None):


        """
        Returns an IdaGeoDataFrame with three columns:
        [
        INDEXERIDA1 : indexer of the first IdaGeoSeries (None if not set),
        INDEXERIDA2 : indexer of the second IdaGeoSeries (None if not set),
        RESULT : the result of the operation
        ]


        Parameters
        ----------
        db2gse_function : str
                Name of the corresponding DB2GSE function.
        valid_types_ida1 : list of str
                Valid input typenames for the first IdaGeoSeries.
        valid_types_ida2 : list of str
                Valid input typenames for the second IdaGeoSeries.
        additional_args : list of str, optional
                Additional arguments for the DB2GSE function.

        Returns
        -------
        IdaGeoDataFrame
        """
        ida1 = self
        
        # Check if allowed data type
        if not (ida1.dtypes.TYPENAME[0] in valid_types_ida1 or
                        valid_types_ida1[0] == 'ST_GEOMETRY'):
            raise TypeError("Column " + ida1.column +
                            " has incompatible type.")
        if not (ida2.dtypes.TYPENAME[0] in valid_types_ida2 or
                        valid_types_ida2[0] == 'ST_GEOMETRY'):
            raise TypeError("Column " + ida2.column +
                            " has incompatible type.")

        # Get the definitions of the columns, which will be the arguments for
        # the DB2GSE function
        column1_for_db2gse = 'IDA1.\"%s\"' %(ida1.geometry.column)
        column2_for_db2gse = 'IDA2.\"%s\"' %(ida2.geometry.column)
        arguments_for_db2gse_function = [column1_for_db2gse, column2_for_db2gse]
        if additional_args is not None:
            for arg in additional_args:
                arguments_for_db2gse_function.append(arg)

        # SELECT statement
        select_columns=[]
        if hasattr(ida1, '_indexer') and ida1._indexer is not None:
            select_columns.append('IDA1.\"%s\" AS \"INDEXERIDA1\"' %(ida1.indexer))
        else:
            message = (ida1+"has no indexer defined. Please assign index column with set_indexer and retry.")
            raise IdaGeoDataFrameError(message)
        if hasattr(ida2, '_indexer') and ida2._indexer is not None:
            select_columns.append('IDA2.\"%s\" AS \"INDEXERIDA2\"' %(ida2.indexer))
        else:
            message = (ida2+"has no indexer defined. Please assign index column with set_indexer and retry.")
            raise IdaGeoDataFrameError(message)
        result_column = (
            db2gse_function+
            '('+
            ','.join(map(str, arguments_for_db2gse_function))+
            ')'
        )
        select_columns.append('%s AS \"RESULT\"' %(result_column))
        select_statement = 'SELECT '+','.join(select_columns)+' '        
        
        # FROM clause
        from_clause=(
            'FROM '+
            ida1.name+' AS IDA1, '+
            ida2.name+' AS IDA2 '
        )

        # Create a view
        view_creation_query='('+select_statement+from_clause+')'
        viewname=self._idadb._create_view_from_expression(view_creation_query)

        idageodf=ibmdbpy4nps.IdaGeoDataFrame(self._idadb, viewname, indexer='INDEXERIDA1')
        return idageodf


import pytest
from ibmdbpy4nps import IdaDataBase, IdaDataFrame
from ibmdbpy4nps.ae import NZFunTApply, NZFunApply, NZFunGroupedApply
from ibmdbpy4nps.ae import NZInstall
import pandas as pd


def test_tapply_host_weather_train_pred():
    idadb = IdaDataBase('weather', 'admin', 'password', verbose=True)
    print(idadb)

    idadf = IdaDataFrame(idadb, 'WEATHER')


    code_str_host = """def decision_tree_ml_host(self, df):

        from sklearn.model_selection import cross_val_score
        from sklearn.impute import SimpleImputer
        from sklearn.tree import DecisionTreeClassifier
        from sklearn.model_selection import train_test_split

        from sklearn.preprocessing import LabelEncoder
        import numpy as np

        result = df.groupby('LOCATION')
        #result = df.groupby(pd.qcut(df['ID'], q=3))

        # result = idadf.ida_query(query, autocommit=True)

        for name, group in result:
            # print(name)

            # print(group)
            def decision_tree_classifier(df):
                imputed_df = df.copy()
                ds_size = len(imputed_df)
                temp_dict = dict()





                columns = imputed_df.columns

                for column in columns:
                    if column=='ID':
                        continue

                    if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
                        if imputed_df[column].isnull().sum()==len(imputed_df):
                         imputed_df[column] = imputed_df[column].fillna(0)

                        else :
                         imp = SimpleImputer(missing_values=np.nan, strategy='mean')
                         transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                         imputed_df[column] = transformed_column


                    if (imputed_df[column].dtype == 'object'):
                        # impute missing values for categorical variables
                        imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
                        imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                        imputed_df[column] = imputed_df[column].astype('str')
                        le = LabelEncoder()
                        # print(imputed_df[column].unique())

                        le.fit(imputed_df[column])
                        # print(le.classes_)
                        imputed_df[column] = le.transform(imputed_df[column])
                        temp_dict[column] = le

                X = imputed_df.drop(['RISK_MM', 'RAINTOMORROW'], axis=1)
                y = imputed_df['RAINTOMORROW']
                X_train, X_test, y_train, y_test = train_test_split(X,y, test_size = 0.25, random_state=42, stratify=y)
                #X_train_mod = X_train.drop(['RISK_MM'],axis=1)
                #X_test_mod = X_test.drop(['RISK_MM'],axis=1)
                dt = DecisionTreeClassifier(max_depth=5)
                dt.fit(X_train, y_train)

                accuracy = dt.score(X_test, y_test)    
                #print(accuracy)



                pred_df = X_test.copy()


                y_pred= dt.predict(X_test)

                pred_df['RAINTOMORROW'] = y_pred
                pred_df['DATASET_SIZE'] = ds_size
                pred_df['CLASSIFIER_ACCURACY']=round(accuracy,2)




                original_columns = pred_df.columns

                for column in original_columns:

                 if column in temp_dict:   
                   pred_df[column] = temp_dict[column].inverse_transform(pred_df[column])
                   #print(pred_df)

                def print_output(x):
                    row = [x['ID'], x['RAINTOMORROW'], x['DATASET_SIZE'], x['CLASSIFIER_ACCURACY']]
                    self.output(row)


                pred_df.apply(print_output, axis=1)

                return pred_df


            ml_result = decision_tree_classifier(df=group)

            """



    output_signature = {'ID': 'int', 'RAINTOMORROW_PRED': 'str', 'DATASET_SIZE': 'int', 'CLASSIFIER_ACCURACY': 'float'}

    import time
    start = time.time()

    nz_tapply = NZFunTApply(df=idadf, code_str=code_str_host, fun_name='decision_tree_ml_host', parallel=False,
                            output_signature=output_signature, merge_output_with_df=True)

    result = nz_tapply.get_result()
    print("\n")
    print(result)
    end = time.time()
    print(end - start)
    result = result.as_dataframe()
    print(result)
    print(result.shape[0])
    assert result.shape[0] == 35565, "number of records are not matching"
    assert result.shape[1] == 28, "number of columns are not matching"


def test_apply_weather_save_table_merge_withdf():
    idadb = IdaDataBase('weather', 'admin', 'password', verbose=True)


    idadf = IdaDataFrame(idadb, 'WEATHER')
    code_str_apply = """def apply_fun(self, x):
        from math import sqrt
        max_temp = x[3]
        id = x[24]
        fahren_max_temp = (max_temp*1.8)+32
        row = [id, max_temp,  fahren_max_temp]
        self.output(row)
        """
    output_signature = {'ID': 'int', 'MAX_TEMP': 'float', 'FAHREN_MAX_TEMP': 'float'}
    output_table = "temp_conversion"
    if idadb.exists_table(output_table):
        idadb.drop_table(output_table)

    nz_apply = NZFunApply(df=idadf, code_str=code_str_apply, fun_name='apply_fun', output_table=output_table,
                          output_signature=output_signature, merge_output_with_df=True)
    result = nz_apply.get_result()
    result = result.as_dataframe()
    print(result)
    assert result.shape[0] == 142193, "number of records are not matching"
    assert result.shape[1] == 27, "number of columns are not matching"


def test_summary():
    idadb = IdaDataBase('weather', 'admin', 'password')
    print(idadb)
    idadf = IdaDataFrame(idadb, 'WEATHER')
    result = idadf.summary()
    print(result)
    assert result.shape[0] == 25, "number of rows are not matching"
    assert result.shape[1] == 13, "number of columns are not matching"

def test_corr():
    idadb = IdaDataBase('weather', 'admin', 'password')
    print(idadb)
    idadf = IdaDataFrame(idadb, 'WEATHER')
    result_df = idadf.corr()

    print(result_df)
    assert result_df.shape[0] == 16, "number of rows are not matching"
    assert result_df.shape[1] == 16, "number of columns are not matching"

def test_cov():
    idadb = IdaDataBase('weather', 'admin', 'password')
    print(idadb)
    idadf = IdaDataFrame(idadb, 'WEATHER')
    result_df = idadf.cov()

    print(result_df)
    assert result_df.shape[0] == 16, "number of rows are not matching"
    assert result_df.shape[1] == 16, "number of columns are not matching"

def test_apply_weather_merge_withdf():
        idadb = IdaDataBase('weather', 'admin', 'password')


        idadf = IdaDataFrame(idadb, 'WEATHER')
        code_str_apply = """def apply_fun(self, x):
            from math import sqrt
            max_temp = x[3]
            id = x[24]
            fahren_max_temp = (max_temp*1.8)+32
            row = [id, max_temp,  fahren_max_temp]
            self.output(row)
            """
        output_signature = {'ID': 'int', 'MAX_TEMP': 'float', 'FAHREN_MAX_TEMP': 'float'}
        nz_apply = NZFunApply(df=idadf, code_str=code_str_apply, fun_name='apply_fun',
                              output_signature=output_signature, merge_output_with_df=True)
        result = nz_apply.get_result()
        print(result)
        assert result.shape[0] == 142193, "number of records are not matching"
        assert result.shape[1] == 27, "number of columns are not matching"

def test_apply_weather_funstr():
            idadb = IdaDataBase('weather', 'admin', 'password', verbose=True)
            print(idadb)

            idadf = IdaDataFrame(idadb, 'WEATHER')
            code_str_apply = """def apply_fun(self, x):
                from math import sqrt
                max_temp = x[3]
                id = x[24]
                fahren_max_temp = (max_temp*1.8)+32
                row = [id, max_temp,  fahren_max_temp]
                self.output(row)
                """
            output_signature = {'ID': 'int', 'MAX_TEMP': 'float', 'FAHREN_MAX_TEMP': 'float'}
            nz_apply = NZFunApply(df=idadf, code_str=code_str_apply, fun_name='apply_fun',
                                  output_signature=output_signature)
            result = nz_apply.get_result()
            print(result)
            assert result.shape[0] == 142193, "number of records are not matching"
            assert result.shape[1] == 3, "number of columns are not matching"


def test_apply_weather_funref():
    idadb = IdaDataBase('weather', 'admin', 'password')
    print(idadb)

    idadf = IdaDataFrame(idadb, 'WEATHER')
    def apply_fun(self, x):
                from math import sqrt
                max_temp = x[3]
                id = x[24]
                fahren_max_temp = (max_temp*1.8)+32
                row = [id, max_temp,  fahren_max_temp]
                self.output(row)

    output_signature = {'ID': 'int', 'MAX_TEMP': 'float', 'FAHREN_MAX_TEMP': 'float'}
    nz_apply = NZFunApply(df=idadf, fun_ref=apply_fun,
                          output_signature=output_signature)
    result = nz_apply.get_result()
    print(result)
    assert result.shape[0] == 142193, "number of records are not matching"
    assert result.shape[1] == 3, "number of columns are not matching"


def test_tapply_weather_host_spus_train_pred():
 idadb = IdaDataBase('weather', 'admin', 'password')
 print(idadb)

 idadf = IdaDataFrame(idadb, 'WEATHER')

 code_str_host_spus = """def decision_tree_ml(self, df):
             from sklearn.model_selection import cross_val_score
             from sklearn.impute import SimpleImputer
             from sklearn.tree import DecisionTreeClassifier
             from sklearn.model_selection import train_test_split

             from sklearn.preprocessing import LabelEncoder
             import numpy as np



             # data preparation
             imputed_df = df.copy()
             ds_size = len(imputed_df)
             temp_dict = dict()


             columns = imputed_df.columns

             for column in columns:
                 if column=='ID':
                     continue

                 if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
                   if imputed_df[column].isnull().sum()==len(imputed_df):
                      imputed_df[column] = imputed_df[column].fillna(0)

                   else :

                      imp = SimpleImputer(missing_values=np.nan, strategy='mean')
                      transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))         
                      imputed_df[column] = transformed_column

                 if (imputed_df[column].dtype == 'object'):
                     # impute missing values for categorical variables
                     imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
                     imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                     imputed_df[column] = imputed_df[column].astype('str')
                     le = LabelEncoder()

                     le.fit(imputed_df[column])
                     # print(le.classes_)
                     imputed_df[column] = le.transform(imputed_df[column])
                     temp_dict[column] = le



             # Create a decision tree
             dt = DecisionTreeClassifier(max_depth=5)
             X = imputed_df.drop(['RISK_MM', 'RAINTOMORROW'], axis=1)
             y = imputed_df['RAINTOMORROW']
             X_train, X_test, y_train, y_test = train_test_split(X,y, test_size = 0.25, random_state=42, stratify=y)


             dt.fit(X_train, y_train)

             accuracy = dt.score(X_test, y_test)    
             #print(accuracy)



             pred_df = X_test.copy()


             y_pred= dt.predict(X_test)

             pred_df['RAINTOMORROW'] = y_pred
             pred_df['DATASET_SIZE'] = ds_size
             pred_df['CLASSIFIER_ACCURACY']=round(accuracy,2)




             original_columns = pred_df.columns

             for column in original_columns:

              if column in temp_dict:   
                pred_df[column] = temp_dict[column].inverse_transform(pred_df[column])
                #print(pred_df)

             def print_output(x):
                 row = [x['ID'], x['RAINTOMORROW'], x['DATASET_SIZE'], x['CLASSIFIER_ACCURACY']]
                 self.output(row)


             pred_df.apply(print_output, axis=1)






 """

 output_signature = {'ID': 'int', 'RAINTOMORROW_PRED': 'str', 'DATASET_SIZE': 'int', 'CLASSIFIER_ACCURACY': 'float'}
 nz_tapply = NZFunTApply(df=idadf, code_str=code_str_host_spus, fun_name ="decision_tree_ml", parallel=True, output_signature=output_signature)
 result = nz_tapply.get_result()
 print("Host +SPUs execution - slicing on a default column- ML function for the entire slices")
 print(result)
 assert result.shape[0] == 35551, "number of records are not matching"
 assert result.shape[1] == 4, "number of columns are not matching"





def test_groupedapply_weather_host_spus():
    idadb = IdaDataBase('weather', 'admin', 'password')
    print(idadb)

    idadf = IdaDataFrame(idadb, 'WEATHER')

    code_str_host_spus = """def decision_tree_ml(self, df):
                from sklearn.model_selection import cross_val_score
                from sklearn.impute import SimpleImputer
                from sklearn.tree import DecisionTreeClassifier
                from sklearn.model_selection import train_test_split

                from sklearn.preprocessing import LabelEncoder
                import numpy as np



                # data preparation
                imputed_df = df.copy()
                ds_size = len(imputed_df)
                temp_dict = dict()


                columns = imputed_df.columns

                for column in columns:
                    if column=='ID':
                        continue

                    if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
                      if imputed_df[column].isnull().sum()==len(imputed_df):
                         imputed_df[column] = imputed_df[column].fillna(0)

                      else :

                         imp = SimpleImputer(missing_values=np.nan, strategy='mean')
                         transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))         
                         imputed_df[column] = transformed_column

                    if (imputed_df[column].dtype == 'object'):
                        # impute missing values for categorical variables
                        imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
                        imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                        imputed_df[column] = imputed_df[column].astype('str')
                        le = LabelEncoder()

                        le.fit(imputed_df[column])
                        # print(le.classes_)
                        imputed_df[column] = le.transform(imputed_df[column])
                        temp_dict[column] = le



                # Create a decision tree
                dt = DecisionTreeClassifier(max_depth=5)
                X = imputed_df.drop(['RISK_MM', 'RAINTOMORROW'], axis=1)
                y = imputed_df['RAINTOMORROW']
                X_train, X_test, y_train, y_test = train_test_split(X,y, test_size = 0.25, random_state=42, stratify=y)


                dt.fit(X_train, y_train)

                accuracy = dt.score(X_test, y_test)    
                #print(accuracy)



                pred_df = X_test.copy()


                y_pred= dt.predict(X_test)

                pred_df['RAINTOMORROW'] = y_pred
                pred_df['DATASET_SIZE'] = ds_size
                pred_df['CLASSIFIER_ACCURACY']=round(accuracy,2)




                original_columns = pred_df.columns

                for column in original_columns:

                 if column in temp_dict:   
                   pred_df[column] = temp_dict[column].inverse_transform(pred_df[column])
                   #print(pred_df)

                def print_output(x):
                    row = [x['ID'], x['RAINTOMORROW'], x['DATASET_SIZE'], x['CLASSIFIER_ACCURACY']]
                    self.output(row)


                pred_df.apply(print_output, axis=1)






    """

    output_signature = {'ID': 'int', 'RAINTOMORROW_PRED': 'str', 'DATASET_SIZE': 'int', 'CLASSIFIER_ACCURACY': 'float'}
    import time
    start = time.time()

    nz_groupapply = NZFunGroupedApply(df=idadf, code_str=code_str_host_spus, index='LOCATION',
                                      fun_name="decision_tree_ml", output_signature=output_signature,
                                      merge_output_with_df=True)

    result = nz_groupapply.get_result()
    result = result.as_dataframe()
    print("Host+ SPUs execution - slicing on user selection -ML function for partitions within slices\n")
    print(result)
    print(time.time()-start)
    assert result.shape[0] == 35565, "number of records are not matching"
    assert result.shape[1] == 28, "number of columns are not matching"

def test_groupedapply_weather_host_spus_funref():
    idadb = IdaDataBase('weather', 'admin', 'password')
    print(idadb)

    idadf = IdaDataFrame(idadb, 'WEATHER')

    def decision_tree_ml(self, df):
                from sklearn.model_selection import cross_val_score
                from sklearn.impute import SimpleImputer
                from sklearn.tree import DecisionTreeClassifier
                from sklearn.model_selection import train_test_split

                from sklearn.preprocessing import LabelEncoder
                import numpy as np



                # data preparation
                imputed_df = df.copy()
                ds_size = len(imputed_df)
                temp_dict = dict()


                columns = imputed_df.columns

                for column in columns:
                    if column=='ID':
                        continue

                    if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
                      if imputed_df[column].isnull().sum()==len(imputed_df):
                         imputed_df[column] = imputed_df[column].fillna(0)

                      else :

                         imp = SimpleImputer(missing_values=np.nan, strategy='mean')
                         transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                         imputed_df[column] = transformed_column

                    if (imputed_df[column].dtype == 'object'):
                        # impute missing values for categorical variables
                        imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
                        imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                        imputed_df[column] = imputed_df[column].astype('str')
                        le = LabelEncoder()

                        le.fit(imputed_df[column])
                        # print(le.classes_)
                        imputed_df[column] = le.transform(imputed_df[column])
                        temp_dict[column] = le



                # Create a decision tree
                dt = DecisionTreeClassifier(max_depth=5)
                X = imputed_df.drop(['RISK_MM', 'RAINTOMORROW'], axis=1)
                y = imputed_df['RAINTOMORROW']
                X_train, X_test, y_train, y_test = train_test_split(X,y, test_size = 0.25, random_state=42, stratify=y)


                dt.fit(X_train, y_train)

                accuracy = dt.score(X_test, y_test)
                #print(accuracy)



                pred_df = X_test.copy()


                y_pred= dt.predict(X_test)

                pred_df['RAINTOMORROW'] = y_pred
                pred_df['DATASET_SIZE'] = ds_size
                pred_df['CLASSIFIER_ACCURACY']=round(accuracy,2)




                original_columns = pred_df.columns

                for column in original_columns:

                 if column in temp_dict:
                   pred_df[column] = temp_dict[column].inverse_transform(pred_df[column])
                   #print(pred_df)

                def print_output(x):
                    row = [x['ID'], x['RAINTOMORROW'], x['DATASET_SIZE'], x['CLASSIFIER_ACCURACY']]
                    self.output(row)


                pred_df.apply(print_output, axis=1)








    output_signature = {'ID': 'int', 'RAINTOMORROW_PRED': 'str', 'DATASET_SIZE': 'int', 'CLASSIFIER_ACCURACY': 'float'}
    import time
    start = time.time()

    nz_groupapply = NZFunGroupedApply(df=idadf,  index='LOCATION',
                                      fun_ref=decision_tree_ml, output_signature=output_signature,
                                      merge_output_with_df=True)

    result = nz_groupapply.get_result()
    print("Host+ SPUs execution - slicing on user selection -ML function for partitions within slices\n")
    result = result.as_dataframe()
    print(result)
    print(time.time()-start)

    assert result.shape[0] == 35565, "number of records are not matching"
    assert result.shape[1] == 28, "number of columns are not matching"


def test_install():
    nzinstall = NZInstall(package_name='scikit-learn')
    result = nzinstall.getResultCode()
    assert result==0, "installation failed"



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 23 09:02:30 2015

@author: efouche
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

from ibmdbpy4nps.internals import idadf_state
import ibmdbpy4nps
from ibmdbpy4nps.utils import timed

import six

@timed
@idadf_state(force = True)   
def discretize(idadf, columns = None, disc= "em", target = None, bins = None, outtable = None, clear_existing=False):
    """
    Discretize a set of numerical columns from an IdaDataFrame and returns an 
    IdaDataFrame open on the discretized version of the dataset. 
    
    Parameters
    ----------
    idadf : IdaDataFrame
    
    columns : str or list of str, optional
        A column or list of columns to be discretized
    
    disc : "ef", "em", "ew", "ewn" default: "em"
        Discretization method to be used
        
        - ef: Discretization bins of equal frequency 
        
        - em: Discretization bins of minimal entropy 
        
        - ew: Discretization bins of equal width
        
        - ewn: Discretization bins of equal width with human-friendly limits 
    
    target : str
        Target column again which the discretization will be done. Relevant
        only for "em" discretization. 
        
    bins: int, optional
        Number of bins. Not relevant for "em" discretization. 
        
    outtable: str, optional
        The name of the output table where the assigned clusters are stored.
        If this parameter is not specified, it is generated automatically.
        If the parameter corresponds to an existing table in the database,
        it is replaced.
    
    clear_existing: bool, default: False
        If set to True, a table will be replaced when a table with the same 
        name already exists  in the database.
    """
    if columns is None:
        columns = idadf._get_numerical_columns()
        if target is not None:
            columns = [x for x in columns if columns != target]    
    else:
        if isinstance(columns, six.string_types):
                columns = [columns]
                
    stored_proc = _check(idadf, columns, disc, target, bins, outtable)

    bound_outtable = idadf._idadb._get_valid_tablename('DISC_BOUNDS_%s_'%idadf.tablename)
    intable = idadf.name   # either the table or a view on the top 
    incolumn = "\";\"".join(columns)
    
    
    # Calculate bounds
    idadf._idadb._call_stored_procedure(stored_proc,
                                        outtable=bound_outtable,
                                        intable=intable,
                                        incolumn=incolumn,
                                        target=target,
                                        bins=bins)
        
    # Create discretized dataset
        
    if outtable is None:
        disc_outtable = idadf._idadb._get_valid_tablename('DISC_%s_'%idadf.tablename)
    else:
        if clear_existing is True:
            try:
                idadf._idadb.drop_table(outtable)
            except:
                pass
        disc_outtable = outtable
    
    try:
        idadf._idadb._call_stored_procedure("APPLY_DISC",
                                            outtable=disc_outtable,
                                            intable=intable,
                                            btable=bound_outtable,
                                            replace="T")
    except:
        raise
    finally:
        idadf._idadb.drop_table(bound_outtable)
    
    return ibmdbpy4nps.IdaDataFrame(idadf._idadb, disc_outtable)
    
    
def _check(idadf, columns, disc, target, bins, outtable):
    """
    Helper function to handle basic checks for 
    ibmdbpy.feature_selection.discretize
    """
    if outtable is not None:
        ibmdbpy4nps.utils.check_tablename(outtable)
    if bins is not None:
        if not isinstance(bins, int):
            raise TypeError("bins argument is not of integer type")
            
    if columns is not None:
        
        if target is not None:
            if target in columns:
                raise ValueError("Target in columns.")
        unknown = []
        for column in columns:
            if column not in idadf.columns:
                unknown.append(column)
        if unknown:
            raise ValueError("Undefined columns: %s"%", ".join(unknown))
        
    if disc == "em":
        if bins is not None:
            raise ValueError("Number of bins is automatically detected for Entropy Minimization discretization.")
        if target is None:
            raise ValueError("Need to define a target for Entropy Minimization discretization.")
        if target in columns:
            raise ValueError("Target column %s cannot be discretize too"%target)
        if target not in idadf.columns:
            raise ValueError("Undefined target column %s"%target)
        stored_proc = "EMDISC"
    else:
        if target is not None:
            raise ValueError("Target attribute defined only for Entropy Minimization discretization.")
        if bins is None:
            bins = 10      
        if disc == "ef":
            stored_proc = "EFDISC"
        elif disc == "ew":
            stored_proc = "EWDISC"
        elif disc == "ewn":
            stored_proc = "EWDISC_NICE"
        else:
            raise ValueError("Unknown discretization method.")
            
    return stored_proc


from ibmdbpy4nps import IdaDataFrame


def build_result(output_table, merge_output, db, df, output_signature, table_name, query, id):
    if output_table and merge_output is False:
        if db.exists_table(output_table):
            raise NameError("table name already exists..choose a different name")

        create_string = "create table " + output_table + " as "
        query = create_string + query
        result = df.ida_query(query, autocommit=True)
        idadf = IdaDataFrame(db, output_table)
        return idadf
    if output_table and merge_output is True:
        if db.exists_table(output_table):
            raise NameError("table name already exists..choose a different name")

        # check for duplicate columns

        for column in output_signature:



            if column in df.columns.tolist():
                if (column == id):
                    continue

                raise ValueError("column " + column + " duplicated in the output table")


        # create a random table for merging the two tables
        output_table_tmp = db._get_valid_tablename(prefix="pyida_")
        create_string = "create table " + output_table_tmp + " as "
        query = create_string + query
        result = df.ida_query(query, autocommit=True)

        # join the two
        columns_str = ""
        for column in output_signature:

            if column == id:
                continue
            columns_str = columns_str + "link." + column + ","
        if len(columns_str) > 0:
            columns_str = columns_str[:-1]

        query = "create table " + output_table + " as  select  " + columns_str + " , base.*  from  " + output_table_tmp + " as link INNER JOIN " \
                                                                                                                             " " + table_name + "  as base on link." + id + " = base." + id + ";"


        result = df.ida_query(query, autocommit=True)
        db.drop_table(output_table_tmp)
        idadf = IdaDataFrame(db, output_table)
        return idadf
    if output_table is None and merge_output is True:
        # create a random table and store query results
        output_table_tmp = db._get_valid_tablename(prefix="pyida_")
        create_string = "create table " + output_table_tmp + " as "
        query = create_string + query
        result = df.ida_query(query, autocommit=True)

        # join the two

        # check if there are duplicate columns in the two tables

        for column in output_signature:

            if column in df.columns.tolist():
                if (column == id):
                    continue
                raise ValueError("column "+column+" duplicated in the output table")

        columns_str = ""
        for column in output_signature:
            if column == id:
                continue
            columns_str = columns_str + "link. " + column + ","
        if len(columns_str) > 0:
            columns_str = columns_str[:-1]
        df_output_table = db._get_valid_tablename(prefix="df_")

        query = "create table "+df_output_table+" as  select  " + columns_str + " , base.*  from  " + output_table_tmp + " as link INNER JOIN " \
                                                                                                                         " " + table_name+"  as base on link." + id + " = base." +id + ";"


        result = df.ida_query(query, autocommit=True)

        idadf = IdaDataFrame(db, df_output_table)
        db.drop_table(output_table_tmp)
        return idadf

    df_output_table = db._get_valid_tablename(prefix="df_")
    query = "create table " + df_output_table + " as " + query
    result = df.ida_query(query)
    idadf = IdaDataFrame(db, df_output_table)
    return idadf



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 23 10:46:25 2015

@author: efouche
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import six

def _check_input(idadf, target, features, ignore_indexer=True):
    """
    Check if the input is valid, i.e. if each column in target and features
    exists in idadf. 
    
    Parameters
    ----------
    target: str or list of str
        A column or list of columns to be used as target
        
    features: str or list of str
        A column or list of columns to be used as feature
    
    ignore_indexer: bool, default: True
        If True, remove the indexer from the features set, as long as an
        indexer is defined in idadf
    """
    #import pdb ; pdb.set_trace()
    if target is not None:
        if isinstance(target, six.string_types):
            if target not in idadf.columns:
                raise ValueError("Unknown target column %s"%target)
            target = [target]
        else:
            if hasattr(target, '__iter__'):
                target = list(target)
            for x in target:
                if x not in idadf.columns:
                    raise ValueError("Unknown target column %s"%x)
            
    if features is not None:
        if isinstance(features, six.string_types):
            if features not in idadf.columns:
                raise ValueError("Unknown feature column %s"%features)
            features = [features]
        else:
            if hasattr(features, '__iter__'):
                features = list(features)
            for x in features:
                if x not in idadf.columns:
                    raise ValueError("Unknown feature column %s"%x)
        if target is None:
            if len(features) == 1:
                raise ValueError("Cannot compute correlation coefficients of only one"+
                                 " column (%s), need at least 2"%features[0])
    else:
        if target is not None:
            if len(target) == 1:
                features = [x for x in idadf.columns if x not in target]
            else:
                features = list(idadf.columns)
        else:
            features = list(idadf.columns)
            
        ## Remove indexer from feature list
        # This is useless and expensive to compute with a primary key 
        if ignore_indexer is True:
            if idadf.indexer:
                if idadf.indexer in features:
                    features.remove(idadf.indexer)
    
    # Catch the case where users ask for the correlation between the two same columns
    #import pdb ; pdb.set_trace()
    if target == features:
        if len(target) == 1:
            raise ValueError("The correlation value of two same columns is always maximal")
            
    if target is None:
        if features is None:
            target = list(idadf.columns) 
        else:
            target = features
            
    return target, features


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

"""The module implement objects and functions that are used to filter IdaDataFrame objects"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import str
from future import standard_library
standard_library.install_aliases()

from copy import deepcopy
from numbers import Number
from ibmdbpy4nps.exceptions import IdaDataBaseError
import six

class FilterQuery(object):
    """
    FilterQueries are used to represent the filtering of an IdaDataFrame
    in ibmdbpy. The use of comparison operators, such as <, <=, ==, >=, >
    on an IdaDataFrame instance produces a FilterQuery instance which acts as a
    container for the where clause of the corresponding SQL request.

    Filtering is possible using a Pandas-like syntax. Applying comparison
    operators to IdaDataFrames produces a FilterQuery instance which
    contains the string embedding the corresponding where clause in the 
    “wherestr” attribute.

    FilterQuery objects also contain a logic that allows them to be
    combined, thus allowing complex filtering.

      You can combine the following operators: \|, \&, \^ (OR, AND and XOR)

    Examples
    --------
    >>> idadf[['sepal_length', 'petal_width'] < 5]
    >>> <ibmdbpy.filtering.FilterQuery at 0xa65ba90>
    >>> _.wherestr
    '("sepal_length" < 5 AND "petal_width" < 5)'

    >>> idadf[idadf[['sepal_length', 'petal_width']] < 5]
    <ibmdbpy.frame.IdaDataFrame at 0xa73a860>
    >>> _.head() # filtered IdaDataFrame
        sepal_length  sepal_width  petal_length  petal_width     species
    0           4.4          2.9           1.4          0.2      setosa
    1           4.7          3.2           1.6          0.2      setosa
    2           4.9          2.5           4.5          1.7   virginica
    3           4.9          2.4           3.3          1.0  versicolor
    4           4.6          3.2           1.4          0.2      setosa

    >>> idadf[(idadf['sepal_length'] < 5) & (idadf[petal_width'] > 1.5)]
    <ibmdbpy.frame.IdaDataFrame at 0xa74b9b0>
    >>> _.head()
       sepal_length  sepal_width  petal_length  petal_width    species
    0           4.9          2.5           4.5          1.7  virginica

    Notes
    -----
    It is not possible to filter an IdaDataFrame by using an IdaDataFrame that 
    is opened in a different data source in the database. This is due to the 
    fact that, using a Pandas-like syntax, “idadf[‘petal_width’] < 5” will 
    return a Boolean array that is used to subset the original DataFrame. This 
    is a fundamental restriction of ibmdbpy: we cannot afford to compute and 
    download such an array because we cannot assume that the result will fit 
    into user’s memory. Download time can also be a performance issue.

    """
    def __init__(self, columns, tablename, method, value):
        """
        Constructor for filterquery objects.

        Parameters
        ----------
        columns : str or list
            columns on which the filter operation should be applied
        tablename : str
            Name of the table to which columns belong
        method : str
            value representing the comparision operator.
            Admissible value:  ["lt","le","eq","ne","ge,"gt"]
        value: str or number
            Value to use to filter.

        Attributes
        ----------
        columns : str or list
            columns on which the filter operation should be applied
        tablename : str
            Name of the table to which columns belong
        method : str
            value representing the comparision operator.
            Admissible value:  ["lt","le","eq","ne","ge,"gt"]
        value: str or number
            Value to use to filter.
        wherestr : str
            SQL where clause to be used for filtering

        Raises
        ------
        IdaDataBaseError
            * The value for filtering is not a string or a number
            * The filtering method is not suppoted
        Notes
        -----
        This object is considered as private, and should be called internally
        by IdaDataFrame instances.
        """
        # Sanity checks
        if isinstance(columns, six.string_types):
            columns = [columns]
        if isinstance(value, six.string_types):
            value = "'" + value + "'"
        if not ((isinstance(value,six.string_types)|isinstance(value,Number))):
            raise IdaDataBaseError("Value for Filterquery is expected to be a string or a number. Type %s"%str(type(value)))
        dictmethod = {"lt": " < ", "le": " <= ",  "eq": " = ", "ne": " != ", "ge": " >= ", "gt": " > "}
        if method not in dictmethod.keys():
            raise IdaDataBaseError("The filtering method is not suppoted. Admissible values for method argument are %s"%list(dictmethod.keys()))

        self.columns = columns     # TO DEPRECATE
        self.tablename = tablename
        self.method = method       # TO DEPRECATE
        self.value = value         # TO DEPRECATE
        self.wherestr = ("(\"" + ("\"" + str(dictmethod[method]) + str(value) + " AND \"").join(columns) +
                        "\"" +  str(dictmethod[method]) + str(value) + ")")

    # TO DEPRECATE
    @property
    def query(self):
        """
        Return an SQL query like "SELECT * FROM %s WHERE <WHERESTR>", where
        <WHERESTR> is the value of the attribute "wherestr".
        """
        return "SELECT * FROM %s WHERE " + self.wherestr

    def __and__(self, other):
        """
        Combine two FilterQuery instances with the operator "&" (AND)
        For example : (idadf['sepal_length'] < 5) & (idadf[petal_width'] > 3)
        """
        self._combine_check(other)
        newquery = deepcopy(self)
        newquery.wherestr = "(%s AND %s)"%(self.wherestr,other.wherestr)
        return newquery

    def __or__(self, other):
        """
        Combine two FilterQuery instances with the operator "|" (OR)
        For example : (idadf['sepal_length'] < 5) | (idadf[petal_width'] > 3)
        """
        self._combine_check(other)
        newquery = deepcopy(self)
        newquery.wherestr =  "(%s OR %s)"%(self.wherestr,other.wherestr)
        return newquery

    def __xor__(self, other):
        """
        Combine two FilterQuery instances with the operator "^" (XOR)
        For example : (idadf['sepal_length'] < 5) ^ (idadf[petal_width'] > 3)
        """
        self._combine_check(other)
        newquery = deepcopy(self)
        newquery.wherestr = ("((NOT %s AND %s) OR (%s AND NOT %s))"%(self.wherestr,other.wherestr,self.wherestr,other.wherestr))
        return newquery

    def _combine_check(self, other):
        """
        Check if the name of two  FilterQuery is the same. Raise an
        IdaDataBaseError in case it is different.
        This function is used before performing logical operations between
        FilterQuery instances.

        Raises
        ------
        IdaDataBaseError
        """
        if self.tablename != other.tablename:
            raise IdaDataBaseError("Combining filtering criterions from "+
                                   "columns belongings to different tables "+
                                   "is not possible.")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

"""
Custom exceptions module
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import str
from future import standard_library
standard_library.install_aliases()

#----------------------------------------------------------------------
# Custom Exception Classes

class Error(Exception):
    """
    This is the base class of all other exceptions thrown by ibmdbpy. It can 
    be used to catch all exceptions with a single except statement.
    """
    def __init__(self, message):
        """
        This is the constructor which take one string argument.
        """
        self._message = message
    def __str__(self):
        """Converts the message to a string."""
        return('ibmdbpy::'+str(self.__class__.__name__)+': '+str(self._message))

class IdaDataBaseError(Error):
    """
    This exception is raised when an error occurs while you manipulate the IdaDataBase interface. 
    """
    pass

class IdaDataFrameError(Error):
    """
    This exception is raised when an error occurs while you manipulate the IdaDataFrame interface. 
    """
    pass

class PrimaryKeyError(Error):
    """
    This exception is raised when an error occurs because of a missing or
    eroneous primary key column.
    """
    pass

class IdaKMeansError(Error):
    """
    This exception is raised when an error related to the ibmdbpy KMeans
    clustering occurs.
    """
    pass

class IdaAssociationRulesError(Error):
    """
    This exception is raised when an error related to ibmdbpy Association Rules occurs.
    """
    pass

class IdaNaiveBayesError(Error):
    """
    This exception is raised when an error related to ibmdbpy Naive Bayes occurs.
    """
    pass

class IdaGeoDataFrameError(Error):
    """
    This exception is raised when an error occurs while you manipulate the IdaDataFrame
    """